import pytest

from geopy.geocoders import OpenCage
from test.geocoders.util import BaseTestGeocoder, env


class TestUnitOpenCage:

    def test_user_agent_custom(self):
        geocoder = OpenCage(
            api_key='DUMMYKEY1234',
            user_agent='my_user_agent/1.0'
        )
        assert geocoder.headers['User-Agent'] == 'my_user_agent/1.0'


@pytest.mark.skipif(
    not bool(env.get('OPENCAGE_KEY')),
    reason="No OPENCAGE_KEY env variables set"
)
class TestOpenCage(BaseTestGeocoder):

    @classmethod
    def make_geocoder(cls, **kwargs):
        return OpenCage(
            api_key=env['OPENCAGE_KEY'],
            timeout=10,
            **kwargs
        )

    async def test_geocode(self):
        await self.geocode_run(
            {"query": "435 north michigan ave, chicago il 60611 usa"},
            {"latitude": 41.890, "longitude": -87.624},
        )

    async def test_unicode_name(self):
        await self.geocode_run(
            {"query": "\u6545\u5bab"},
            {"latitude": 39.916, "longitude": 116.390},
        )

    async def test_geocode_empty_result(self):
        await self.geocode_run(
            {"query": "xqj37"},
            {},
            expect_failure=True
        )

    async def test_bounds(self):
        await self.geocode_run(
            {"query": "moscow",  # Idaho USA
             "bounds": [[50.1, -130.1], [44.1, -100.9]]},
            {"latitude": 46.7323875, "longitude": -117.0001651},
        )

    async def test_country_str(self):
        await self.geocode_run(
            {"query": "kazan",
             "country": 'tr'},
            {"latitude": 40.2317, "longitude": 32.6839},
        )

    async def test_country_list(self):
        await self.geocode_run(
            {"query": "kazan",
             "country": ['cn', 'tr']},
            {"latitude": 40.2317, "longitude": 32.6839},
        )



import pytest

from geopy.exc import GeocoderQueryError
from geopy.geocoders import DataBC
from test.geocoders.util import BaseTestGeocoder


class TestDataBC(BaseTestGeocoder):

    @classmethod
    def make_geocoder(cls, **kwargs):
        return DataBC(**kwargs)

    async def test_user_agent_custom(self):
        geocoder = DataBC(
            user_agent='my_user_agent/1.0'
        )
        assert geocoder.headers['User-Agent'] == 'my_user_agent/1.0'

    async def test_geocode(self):
        await self.geocode_run(
            {"query": "135 North Pym Road, Parksville"},
            {"latitude": 49.321, "longitude": -124.337},
        )

    async def test_unicode_name(self):
        await self.geocode_run(
            {"query": "Barri\u00e8re"},
            {"latitude": 51.179, "longitude": -120.123},
        )

    async def test_multiple_results(self):
        res = await self.geocode_run(
            {"query": "1st St", "exactly_one": False},
            {},
        )
        assert len(res) > 1

    async def test_optional_params(self):
        await self.geocode_run(
            {"query": "5670 malibu terrace nanaimo bc",
             "location_descriptor": "accessPoint",
             "set_back": 100},
            {"latitude": 49.2299, "longitude": -124.0163},
        )

    async def test_query_error(self):
        with pytest.raises(GeocoderQueryError):
            await self.geocode_run(
                {"query": "1 Main St, Vancouver",
                 "location_descriptor": "access_Point"},
                {},
                expect_failure=True,
            )



import asyncio
from unittest.mock import MagicMock, patch, sentinel

import pytest

from geopy.exc import GeocoderQuotaExceeded, GeocoderServiceError
from geopy.extra.rate_limiter import AsyncRateLimiter, RateLimiter


@pytest.fixture(params=[False, True])
def is_async(request):
    return request.param


@pytest.fixture
def auto_async(is_async):
    if is_async:
        async def auto_async(coro):
            return await coro
    else:
        async def auto_async(result):
            return result
    return auto_async


@pytest.fixture
def auto_async_side_effect(is_async):
    def auto_async_side_effect(side_effect):
        if not is_async:
            return side_effect

        mock = MagicMock(side_effect=side_effect)

        async def func(*args, **kwargs):
            return mock(*args, **kwargs)

        return func

    return auto_async_side_effect


@pytest.fixture
def rate_limiter_cls(is_async):
    if is_async:
        return AsyncRateLimiter
    else:
        return RateLimiter


@pytest.fixture
def mock_clock(rate_limiter_cls):
    with patch.object(rate_limiter_cls, '_clock') as mock_clock:
        yield mock_clock


@pytest.fixture
def mock_sleep(auto_async_side_effect, rate_limiter_cls):
    with patch.object(rate_limiter_cls, '_sleep') as mock_sleep:
        mock_sleep.side_effect = auto_async_side_effect(None)
        yield mock_sleep


async def test_min_delay(
    rate_limiter_cls, mock_clock, mock_sleep, auto_async_side_effect, auto_async
):
    mock_func = MagicMock()
    mock_func.side_effect = auto_async_side_effect(None)
    min_delay = 3.5

    mock_clock.side_effect = [1]
    rl = rate_limiter_cls(mock_func, min_delay_seconds=min_delay)

    # First call -- no delay
    clock_first = 10
    mock_clock.side_effect = [clock_first, clock_first]  # no delay here
    await auto_async(rl(sentinel.arg, kwa=sentinel.kwa))
    mock_sleep.assert_not_called()
    mock_func.assert_called_once_with(sentinel.arg, kwa=sentinel.kwa)

    # Second call after min_delay/3 seconds -- should be delayed
    clock_second = clock_first + (min_delay / 3)
    mock_clock.side_effect = [clock_second, clock_first + min_delay]
    await auto_async(rl(sentinel.arg, kwa=sentinel.kwa))
    mock_sleep.assert_called_with(min_delay - (clock_second - clock_first))
    mock_sleep.reset_mock()

    # Third call after min_delay*2 seconds -- no delay again
    clock_third = clock_first + min_delay + min_delay * 2
    mock_clock.side_effect = [clock_third, clock_third]
    await auto_async(rl(sentinel.arg, kwa=sentinel.kwa))
    mock_sleep.assert_not_called()


async def test_max_retries(
    rate_limiter_cls, mock_clock, mock_sleep, auto_async_side_effect, auto_async
):
    mock_func = MagicMock()
    mock_clock.return_value = 1
    rl = rate_limiter_cls(
        mock_func, max_retries=3,
        return_value_on_exception=sentinel.return_value,
    )

    # Non-geopy errors must not be swallowed
    mock_func.side_effect = auto_async_side_effect(ValueError)
    with pytest.raises(ValueError):
        await auto_async(rl(sentinel.arg))
    assert 1 == mock_func.call_count
    mock_func.reset_mock()

    # geopy errors must be swallowed and retried
    mock_func.side_effect = auto_async_side_effect(GeocoderServiceError)
    assert sentinel.return_value == await auto_async(rl(sentinel.arg))
    assert 4 == mock_func.call_count
    mock_func.reset_mock()

    # Successful value must be returned
    mock_func.side_effect = auto_async_side_effect(side_effect=[
        GeocoderServiceError, GeocoderServiceError, sentinel.good
    ])
    assert sentinel.good == await auto_async(rl(sentinel.arg))
    assert 3 == mock_func.call_count
    mock_func.reset_mock()

    # When swallowing is disabled, the exception must be raised
    rl.swallow_exceptions = False
    mock_func.side_effect = auto_async_side_effect(GeocoderQuotaExceeded)
    with pytest.raises(GeocoderQuotaExceeded):
        await auto_async(rl(sentinel.arg))
    assert 4 == mock_func.call_count
    mock_func.reset_mock()


async def test_error_wait_seconds(
    rate_limiter_cls, mock_clock, mock_sleep, auto_async_side_effect, auto_async
):
    mock_func = MagicMock()
    error_wait = 3.3

    mock_clock.return_value = 1
    rl = rate_limiter_cls(
        mock_func, max_retries=3,
        error_wait_seconds=error_wait,
        return_value_on_exception=sentinel.return_value,
    )

    mock_func.side_effect = auto_async_side_effect(GeocoderServiceError)
    assert sentinel.return_value == await auto_async(rl(sentinel.arg))
    assert 4 == mock_func.call_count
    assert 3 == mock_sleep.call_count
    mock_sleep.assert_called_with(error_wait)
    mock_func.reset_mock()


async def test_sync_raises_for_awaitable():
    def g():  # non-async function returning an awaitable -- like `geocode`.
        async def coro():
            pass  # pragma: no cover

        # Make a task from the coroutine, to avoid
        # the `coroutine 'coro' was never awaited` warning:
        task = asyncio.ensure_future(coro())
        return task

    rl = RateLimiter(g)

    with pytest.raises(ValueError):
        await rl()



import pytest

from geopy import exc
from geopy.geocoders import ArcGIS
from geopy.point import Point
from test.geocoders.util import BaseTestGeocoder, env


class TestUnitArcGIS:

    def test_user_agent_custom(self):
        geocoder = ArcGIS(
            user_agent='my_user_agent/1.0'
        )
        assert geocoder.headers['User-Agent'] == 'my_user_agent/1.0'


class TestArcGIS(BaseTestGeocoder):

    @classmethod
    def make_geocoder(cls, **kwargs):
        return ArcGIS(timeout=3, **kwargs)

    async def test_missing_password_error(self):
        with pytest.raises(exc.ConfigurationError):
            ArcGIS(username='a')

    async def test_scheme_config_error(self):
        with pytest.raises(exc.ConfigurationError):
            ArcGIS(
                username='a',
                password='b',
                referer='http://www.example.com',
                scheme='http'
            )

    async def test_geocode(self):
        await self.geocode_run(
            {"query": "435 north michigan ave, chicago il 60611 usa"},
            {"latitude": 41.890, "longitude": -87.624},
        )

    async def test_unicode_name(self):
        await self.geocode_run(
            {"query": "\u6545\u5bab"},
            {"latitude": 39.916, "longitude": 116.390},
        )

    async def test_empty_response(self):
        await self.geocode_run(
            {"query": "dksahdksahdjksahdoufydshf"},
            {},
            expect_failure=True
        )

    async def test_geocode_with_out_fields_string(self):
        result = await self.geocode_run(
            {"query": "Trafalgar Square, London",
             "out_fields": "Country"},
            {}
        )
        assert result.raw['attributes'] == {'Country': 'GBR'}

    async def test_geocode_with_out_fields_list(self):
        result = await self.geocode_run(
            {"query": "Trafalgar Square, London",
             "out_fields": ["City", "Type"]},
            {}
        )
        assert result.raw['attributes'] == {
            'City': 'London', 'Type': 'Tourist Attraction'
        }

    async def test_reverse_point(self):
        location = await self.reverse_run(
            {"query": Point(40.753898, -73.985071)},
            {"latitude": 40.75376406311989, "longitude": -73.98489005863667},
        )
        assert 'New York' in location.address

    async def test_reverse_not_exactly_one(self):
        await self.reverse_run(
            {"query": Point(40.753898, -73.985071), "exactly_one": False},
            {"latitude": 40.75376406311989, "longitude": -73.98489005863667},
        )

    async def test_reverse_no_result(self):
        await self.reverse_run(
            # North Atlantic Ocean
            {"query": (35.173809, -37.485351)},
            {},
            expect_failure=True
        )


@pytest.mark.skipif(
    not (env.get('ARCGIS_USERNAME') is not None
         or env.get('ARCGIS_PASSWORD') is not None
         or env.get('ARCGIS_REFERER') is not None),
    reason="No ARCGIS_USERNAME or ARCGIS_PASSWORD or ARCGIS_REFERER env variable set"
)
class TestArcGISAuthenticated(BaseTestGeocoder):

    @classmethod
    def make_geocoder(cls, **kwargs):
        return ArcGIS(
            username=env['ARCGIS_USERNAME'],
            password=env['ARCGIS_PASSWORD'],
            referer=env['ARCGIS_REFERER'],
            timeout=3,
            **kwargs
        )

    async def test_basic_address(self):
        await self.geocode_run(
            {"query": "Potsdamer Platz, Berlin, Deutschland"},
            {"latitude": 52.5094982, "longitude": 13.3765983},
        )



import pytest

from geopy.geocoders import Geolake
from test.geocoders.util import BaseTestGeocoder, env


class TestUnitGeolake:

    def test_user_agent_custom(self):
        geocoder = Geolake(
            api_key='DUMMYKEY1234',
            user_agent='my_user_agent/1.0'
        )
        assert geocoder.headers['User-Agent'] == 'my_user_agent/1.0'


@pytest.mark.skipif(
    not bool(env.get('GEOLAKE_KEY')),
    reason="No GEOLAKE_KEY env variables set"
)
class TestGeolake(BaseTestGeocoder):

    @classmethod
    def make_geocoder(cls, **kwargs):
        return Geolake(
            api_key=env['GEOLAKE_KEY'],
            timeout=10,
            **kwargs
        )

    async def test_geocode(self):
        await self.geocode_run(
            {"query": "435 north michigan ave, chicago il 60611 usa"},
            {"latitude": 41.890344, "longitude": -87.623234, "address": "Chicago, US"},
        )

    async def test_geocode_country_codes_str(self):
        await self.geocode_run(
            {"query": "Toronto", "country_codes": "US"},
            {"latitude": 40.46, "longitude": -80.6, "address": "Toronto, US"},
        )

    async def test_geocode_country_codes_list(self):
        await self.geocode_run(
            {"query": "Toronto", "country_codes": ["CN", "US"]},
            {"latitude": 40.46, "longitude": -80.6, "address": "Toronto, US"},
        )

    async def test_geocode_structured(self):
        query = {
            "street": "north michigan ave",
            "houseNumber": "435",
            "city": "chicago",
            "state": "il",
            "zipcode": 60611,
            "country": "US"
        }
        await self.geocode_run(
            {"query": query},
            {"latitude": 41.890344, "longitude": -87.623234}
        )

    async def test_geocode_empty_result(self):
        await self.geocode_run(
            {"query": "xqj37"},
            {},
            expect_failure=True
        )

    async def test_geocode_missing_city_in_result(self):
        await self.geocode_run(
            {"query": "H1W 0B4"},
            {"latitude": 45.544952, "longitude": -73.546694, "address": "CA"}
        )



import collections.abc
from functools import partial
from urllib.parse import quote, urlencode

from geopy.exc import (
    GeocoderAuthenticationFailure,
    GeocoderInsufficientPrivileges,
    GeocoderQuotaExceeded,
    GeocoderServiceError,
    GeocoderUnavailable,
)
from geopy.geocoders.base import DEFAULT_SENTINEL, Geocoder
from geopy.location import Location
from geopy.util import join_filter, logger

__all__ = ("Bing", )


class Bing(Geocoder):
    """Geocoder using the Bing Maps Locations API.

    Documentation at:
        https://msdn.microsoft.com/en-us/library/ff701715.aspx
    """

    structured_query_params = {
        'addressLine',
        'locality',
        'adminDistrict',
        'countryRegion',
        'postalCode',
    }

    geocode_path = '/REST/v1/Locations'
    reverse_path = '/REST/v1/Locations/%(point)s'

    def __init__(
            self,
            api_key,
            *,
            scheme=None,
            timeout=DEFAULT_SENTINEL,
            proxies=DEFAULT_SENTINEL,
            user_agent=None,
            ssl_context=DEFAULT_SENTINEL,
            adapter_factory=None
    ):
        """

        :param str api_key: Should be a valid Bing Maps API key
            (https://www.microsoft.com/en-us/maps/create-a-bing-maps-key).

        :param str scheme:
            See :attr:`geopy.geocoders.options.default_scheme`.

        :param int timeout:
            See :attr:`geopy.geocoders.options.default_timeout`.

        :param dict proxies:
            See :attr:`geopy.geocoders.options.default_proxies`.

        :param str user_agent:
            See :attr:`geopy.geocoders.options.default_user_agent`.

        :type ssl_context: :class:`ssl.SSLContext`
        :param ssl_context:
            See :attr:`geopy.geocoders.options.default_ssl_context`.

        :param callable adapter_factory:
            See :attr:`geopy.geocoders.options.default_adapter_factory`.

            .. versionadded:: 2.0
        """
        super().__init__(
            scheme=scheme,
            timeout=timeout,
            proxies=proxies,
            user_agent=user_agent,
            ssl_context=ssl_context,
            adapter_factory=adapter_factory,
        )
        self.api_key = api_key
        domain = 'dev.virtualearth.net'
        self.geocode_api = '%s://%s%s' % (self.scheme, domain, self.geocode_path)
        self.reverse_api = '%s://%s%s' % (self.scheme, domain, self.reverse_path)

    def geocode(
            self,
            query,
            *,
            exactly_one=True,
            user_location=None,
            timeout=DEFAULT_SENTINEL,
            culture=None,
            include_neighborhood=None,
            include_country_code=False
    ):
        """
        Return a location point by address.

        :param str query: The address or query you wish to geocode.

            For a structured query, provide a dictionary whose keys
            are one of: `addressLine`, `locality` (city),
            `adminDistrict` (state), `countryRegion`, or `postalcode`.

        :param bool exactly_one: Return one result or a list of results, if
            available.

        :param user_location: Prioritize results closer to
            this location.
        :type user_location: :class:`geopy.point.Point`

        :param int timeout: Time, in seconds, to wait for the geocoding service
            to respond before raising a :class:`geopy.exc.GeocoderTimedOut`
            exception. Set this only if you wish to override, on this call
            only, the value set during the geocoder's initialization.

        :param str culture: Affects the language of the response,
            must be a two-letter country code.

        :param bool include_neighborhood: Sets whether to include the
            neighborhood field in the response.

        :param bool include_country_code: Sets whether to include the
            two-letter ISO code of the country in the response (field name
            'countryRegionIso2').

        :rtype: ``None``, :class:`geopy.location.Location` or a list of them, if
            ``exactly_one=False``.
        """
        if isinstance(query, collections.abc.Mapping):
            params = {
                key: val
                for key, val
                in query.items()
                if key in self.structured_query_params
            }
            params['key'] = self.api_key
        else:
            params = {
                'query': query,
                'key': self.api_key
            }
        if user_location:
            params['userLocation'] = ",".join(
                (str(user_location.latitude), str(user_location.longitude))
            )
        if exactly_one:
            params['maxResults'] = 1
        if culture:
            params['culture'] = culture
        if include_neighborhood is not None:
            params['includeNeighborhood'] = include_neighborhood
        if include_country_code:
            params['include'] = 'ciso2'  # the only acceptable value

        url = "?".join((self.geocode_api, urlencode(params)))
        logger.debug("%s.geocode: %s", self.__class__.__name__, url)
        callback = partial(self._parse_json, exactly_one=exactly_one)
        return self._call_geocoder(url, callback, timeout=timeout)

    def reverse(
            self,
            query,
            *,
            exactly_one=True,
            timeout=DEFAULT_SENTINEL,
            culture=None,
            include_country_code=False
    ):
        """
        Return an address by location point.

        :param query: The coordinates for which you wish to obtain the
            closest human-readable addresses.
        :type query: :class:`geopy.point.Point`, list or tuple of ``(latitude,
            longitude)``, or string as ``"%(latitude)s, %(longitude)s"``.

        :param bool exactly_one: Return one result or a list of results, if
            available.

        :param int timeout: Time, in seconds, to wait for the geocoding service
            to respond before raising a :class:`geopy.exc.GeocoderTimedOut`
            exception. Set this only if you wish to override, on this call
            only, the value set during the geocoder's initialization.

        :param str culture: Affects the language of the response,
            must be a two-letter country code.

        :param bool include_country_code: Sets whether to include the
            two-letter ISO code of the country in the response (field name
            'countryRegionIso2').

        :rtype: ``None``, :class:`geopy.location.Location` or a list of them, if
            ``exactly_one=False``.
        """
        point = self._coerce_point_to_string(query)
        params = {'key': self.api_key}
        if culture:
            params['culture'] = culture
        if include_country_code:
            params['include'] = 'ciso2'  # the only acceptable value

        quoted_point = quote(point.encode('utf-8'))
        url = "?".join((self.reverse_api % dict(point=quoted_point),
                        urlencode(params)))

        logger.debug("%s.reverse: %s", self.__class__.__name__, url)
        callback = partial(self._parse_json, exactly_one=exactly_one)
        return self._call_geocoder(url, callback, timeout=timeout)

    def _parse_json(self, doc, exactly_one=True):
        """
        Parse a location name, latitude, and longitude from an JSON response.
        """
        status_code = doc.get("statusCode", 200)
        if status_code != 200:
            err = doc.get("errorDetails", "")
            if status_code == 401:
                raise GeocoderAuthenticationFailure(err)
            elif status_code == 403:
                raise GeocoderInsufficientPrivileges(err)
            elif status_code == 429:
                raise GeocoderQuotaExceeded(err)
            elif status_code == 503:
                raise GeocoderUnavailable(err)
            else:
                raise GeocoderServiceError(err)

        resources = doc['resourceSets'][0]['resources']
        if resources is None or not len(resources):
            return None

        def parse_resource(resource):
            """
            Parse each return object.
            """
            stripchars = ", \n"
            addr = resource['address']

            address = addr.get('addressLine', '').strip(stripchars)
            city = addr.get('locality', '').strip(stripchars)
            state = addr.get('adminDistrict', '').strip(stripchars)
            zipcode = addr.get('postalCode', '').strip(stripchars)
            country = addr.get('countryRegion', '').strip(stripchars)

            city_state = join_filter(", ", [city, state])
            place = join_filter(" ", [city_state, zipcode])
            location = join_filter(", ", [address, place, country])

            latitude = resource['point']['coordinates'][0] or None
            longitude = resource['point']['coordinates'][1] or None
            if latitude and longitude:
                latitude = float(latitude)
                longitude = float(longitude)

            return Location(location, (latitude, longitude), resource)

        if exactly_one:
            return parse_resource(resources[0])
        else:
            return [parse_resource(resource) for resource in resources]



from functools import partial
from urllib.parse import quote, urlencode

from geopy.geocoders.base import DEFAULT_SENTINEL, Geocoder
from geopy.location import Location
from geopy.point import Point
from geopy.util import logger

__all__ = ("MapBox", )


class MapBox(Geocoder):
    """Geocoder using the Mapbox API.

    Documentation at:
        https://www.mapbox.com/api-documentation/
    """

    api_path = '/geocoding/v5/mapbox.places/%(query)s.json/'

    def __init__(
            self,
            api_key,
            *,
            scheme=None,
            timeout=DEFAULT_SENTINEL,
            proxies=DEFAULT_SENTINEL,
            user_agent=None,
            ssl_context=DEFAULT_SENTINEL,
            adapter_factory=None,
            domain='api.mapbox.com'
    ):
        """
        :param str api_key: The API key required by Mapbox to perform
            geocoding requests. API keys are managed through Mapox's account
            page (https://www.mapbox.com/account/access-tokens).

        :param str scheme:
            See :attr:`geopy.geocoders.options.default_scheme`.

        :param int timeout:
            See :attr:`geopy.geocoders.options.default_timeout`.

        :param dict proxies:
            See :attr:`geopy.geocoders.options.default_proxies`.

        :param str user_agent:
            See :attr:`geopy.geocoders.options.default_user_agent`.

        :type ssl_context: :class:`ssl.SSLContext`
        :param ssl_context:
            See :attr:`geopy.geocoders.options.default_ssl_context`.

        :param callable adapter_factory:
            See :attr:`geopy.geocoders.options.default_adapter_factory`.

            .. versionadded:: 2.0

        :param str domain: base api domain for mapbox
        """
        super().__init__(
            scheme=scheme,
            timeout=timeout,
            proxies=proxies,
            user_agent=user_agent,
            ssl_context=ssl_context,
            adapter_factory=adapter_factory,
        )
        self.api_key = api_key
        self.domain = domain.strip('/')
        self.api = "%s://%s%s" % (self.scheme, self.domain, self.api_path)

    def _parse_json(self, json, exactly_one=True):
        '''Returns location, (latitude, longitude) from json feed.'''
        features = json['features']
        if features == []:
            return None

        def parse_feature(feature):
            location = feature['place_name']
            longitude = feature['geometry']['coordinates'][0]
            latitude = feature['geometry']['coordinates'][1]
            return Location(location, (latitude, longitude), feature)
        if exactly_one:
            return parse_feature(features[0])
        else:
            return [parse_feature(feature) for feature in features]

    def geocode(
            self,
            query,
            *,
            exactly_one=True,
            timeout=DEFAULT_SENTINEL,
            proximity=None,
            country=None,
            bbox=None
    ):
        """
        Return a location point by address.

        :param str query: The address or query you wish to geocode.

        :param bool exactly_one: Return one result or a list of results, if
            available.

        :param int timeout: Time, in seconds, to wait for the geocoding service
            to respond before raising a :class:`geopy.exc.GeocoderTimedOut`
            exception. Set this only if you wish to override, on this call
            only, the value set during the geocoder's initialization.

        :param proximity: A coordinate to bias local results based on a provided
            location.
        :type proximity: :class:`geopy.point.Point`, list or tuple of ``(latitude,
            longitude)``, or string as ``"%(latitude)s, %(longitude)s"``.

        :param country: Country to filter result in form of
            ISO 3166-1 alpha-2 country code (e.g. ``FR``).
            Might be a Python list of strings.

        :type country: str or list

        :param bbox: The bounding box of the viewport within which
            to bias geocode results more prominently.
            Example: ``[Point(22, 180), Point(-22, -180)]``.
        :type bbox: list or tuple of 2 items of :class:`geopy.point.Point` or
            ``(latitude, longitude)`` or ``"%(latitude)s, %(longitude)s"``.

        :rtype: ``None``, :class:`geopy.location.Location` or a list of them, if
            ``exactly_one=False``.
        """
        params = {}

        params['access_token'] = self.api_key
        if bbox:
            params['bbox'] = self._format_bounding_box(
                bbox, "%(lon1)s,%(lat1)s,%(lon2)s,%(lat2)s")

        if not country:
            country = []
        if isinstance(country, str):
            country = [country]
        if country:
            params['country'] = ",".join(country)

        if proximity:
            p = Point(proximity)
            params['proximity'] = "%s,%s" % (p.longitude, p.latitude)

        quoted_query = quote(query.encode('utf-8'))
        url = "?".join((self.api % dict(query=quoted_query),
                        urlencode(params)))
        logger.debug("%s.geocode: %s", self.__class__.__name__, url)
        callback = partial(self._parse_json, exactly_one=exactly_one)
        return self._call_geocoder(url, callback, timeout=timeout)

    def reverse(
            self,
            query,
            *,
            exactly_one=True,
            timeout=DEFAULT_SENTINEL
    ):
        """
        Return an address by location point.

        :param query: The coordinates for which you wish to obtain the
            closest human-readable addresses.
        :type query: :class:`geopy.point.Point`, list or tuple of ``(latitude,
            longitude)``, or string as ``"%(latitude)s, %(longitude)s"``.

        :param bool exactly_one: Return one result or a list of results, if
            available.

        :param int timeout: Time, in seconds, to wait for the geocoding service
            to respond before raising a :class:`geopy.exc.GeocoderTimedOut`
            exception. Set this only if you wish to override, on this call
            only, the value set during the geocoder's initialization.

        :rtype: ``None``, :class:`geopy.location.Location` or a list of them, if
            ``exactly_one=False``.
        """
        params = {}
        params['access_token'] = self.api_key

        point = self._coerce_point_to_string(query, "%(lon)s,%(lat)s")
        quoted_query = quote(point.encode('utf-8'))
        url = "?".join((self.api % dict(query=quoted_query),
                        urlencode(params)))
        logger.debug("%s.reverse: %s", self.__class__.__name__, url)
        callback = partial(self._parse_json, exactly_one=exactly_one)
        return self._call_geocoder(url, callback, timeout=timeout)



import unittest
from contextlib import ExitStack
from unittest.mock import patch, sentinel

import pytest

import geopy.geocoders
import geopy.geocoders.base
from geopy.adapters import BaseAsyncAdapter, BaseSyncAdapter
from geopy.exc import GeocoderNotFound, GeocoderQueryError
from geopy.geocoders import GoogleV3, get_geocoder_for_service
from geopy.geocoders.base import Geocoder, _synchronized
from geopy.point import Point


class DummySyncAdapter(BaseSyncAdapter):
    def get_json(self, *args, **kwargs):  # pragma: no cover
        raise NotImplementedError

    def get_text(self, *args, **kwargs):  # pragma: no cover
        raise NotImplementedError


class DummyAsyncAdapter(BaseAsyncAdapter):
    async def get_json(self, *args, **kwargs):  # pragma: no cover
        raise NotImplementedError

    async def get_text(self, *args, **kwargs):  # pragma: no cover
        raise NotImplementedError


class GetGeocoderTestCase(unittest.TestCase):

    def test_get_geocoder_for_service(self):
        assert get_geocoder_for_service("google") == GoogleV3
        assert get_geocoder_for_service("googlev3") == GoogleV3

    def test_get_geocoder_for_service_raises_for_unknown(self):
        with pytest.raises(GeocoderNotFound):
            get_geocoder_for_service("")


class GeocoderTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.geocoder = Geocoder()

    def test_init_with_args(self):
        scheme = 'http'
        timeout = 942
        proxies = {'https': '192.0.2.0'}
        user_agent = 'test app'
        ssl_context = sentinel.some_ssl_context
        adapter = DummySyncAdapter(proxies=None, ssl_context=None)

        geocoder = Geocoder(
            scheme=scheme,
            timeout=timeout,
            proxies=proxies,
            user_agent=user_agent,
            ssl_context=ssl_context,
            adapter_factory=lambda **kw: adapter,
        )
        for attr in ('scheme', 'timeout', 'proxies', 'ssl_context'):
            assert locals()[attr] == getattr(geocoder, attr)
        assert user_agent == geocoder.headers['User-Agent']
        assert adapter is geocoder.adapter

    @patch.object(geopy.geocoders.options, 'default_adapter_factory',
                  DummySyncAdapter)
    def test_init_with_defaults(self):
        attr_to_option = {
            'scheme': 'default_scheme',
            'timeout': 'default_timeout',
            'proxies': 'default_proxies',
            'ssl_context': 'default_ssl_context',
        }

        geocoder = Geocoder()
        for geocoder_attr, options_attr in attr_to_option.items():
            assert (
                getattr(geopy.geocoders.options, options_attr) ==
                getattr(geocoder, geocoder_attr)
            )
        assert (
            geopy.geocoders.options.default_user_agent ==
            geocoder.headers['User-Agent']
        )
        assert DummySyncAdapter is type(geocoder.adapter)  # noqa

    @patch.object(geopy.geocoders.options, 'default_proxies', {'https': '192.0.2.0'})
    @patch.object(geopy.geocoders.options, 'default_timeout', 10)
    @patch.object(geopy.geocoders.options, 'default_ssl_context',
                  sentinel.some_ssl_context)
    def test_init_with_none_overrides_default(self):
        geocoder = Geocoder(proxies=None, timeout=None, ssl_context=None)
        assert geocoder.proxies is None
        assert geocoder.timeout is None
        assert geocoder.ssl_context is None

    @patch.object(geopy.geocoders.options, 'default_user_agent',
                  'mocked_user_agent/0.0.0')
    def test_user_agent_default(self):
        geocoder = Geocoder()
        assert geocoder.headers['User-Agent'] == 'mocked_user_agent/0.0.0'

    def test_user_agent_custom(self):
        geocoder = Geocoder(
            user_agent='my_user_agent/1.0'
        )
        assert geocoder.headers['User-Agent'] == 'my_user_agent/1.0'

    @patch.object(geopy.geocoders.options, 'default_timeout', 12)
    def test_call_geocoder_timeout(self):
        url = 'spam://ham/eggs'

        g = Geocoder(adapter_factory=DummySyncAdapter)
        assert g.timeout == 12

        with ExitStack() as stack:
            mock_get_json = stack.enter_context(patch.object(g.adapter, 'get_json'))

            g._call_geocoder(url, lambda res: res)
            args, kwargs = mock_get_json.call_args
            assert kwargs['timeout'] == 12

            g._call_geocoder(url, lambda res: res, timeout=7)
            args, kwargs = mock_get_json.call_args
            assert kwargs['timeout'] == 7

            g._call_geocoder(url, lambda res: res, timeout=None)
            args, kwargs = mock_get_json.call_args
            assert kwargs['timeout'] is None

    def test_ssl_context(self):
        with ExitStack() as stack:
            mock_adapter = stack.enter_context(
                patch.object(
                    geopy.geocoders.base.options,
                    'default_adapter_factory',
                    return_value=DummySyncAdapter(proxies=None, ssl_context=None),
                )
            )

            for ssl_context in (None, sentinel.some_ssl_context):
                Geocoder(ssl_context=ssl_context)
                args, kwargs = mock_adapter.call_args
                assert kwargs['ssl_context'] is ssl_context


class GeocoderPointCoercionTestCase(unittest.TestCase):
    coordinates = (40.74113, -73.989656)
    coordinates_str = "40.74113,-73.989656"
    coordinates_address = "175 5th Avenue, NYC, USA"

    def setUp(self):
        self.method = Geocoder()._coerce_point_to_string

    def test_point(self):
        latlon = self.method(Point(*self.coordinates))
        assert latlon == self.coordinates_str

    def test_tuple_of_floats(self):
        latlon = self.method(self.coordinates)
        assert latlon == self.coordinates_str

    def test_string(self):
        latlon = self.method(self.coordinates_str)
        assert latlon == self.coordinates_str

    def test_string_is_trimmed(self):
        coordinates_str_spaces = "  %s  ,  %s  " % self.coordinates
        latlon = self.method(coordinates_str_spaces)
        assert latlon == self.coordinates_str

    def test_output_format_is_respected(self):
        expected = "  %s  %s  " % self.coordinates[::-1]
        lonlat = self.method(self.coordinates_str, "  %(lon)s  %(lat)s  ")
        assert lonlat == expected

    def test_address(self):
        with pytest.raises(ValueError):
            self.method(self.coordinates_address)


class GeocoderFormatBoundingBoxTestCase(unittest.TestCase):

    def setUp(self):
        self.method = Geocoder()._format_bounding_box

    def test_string_raises(self):
        with pytest.raises(GeocoderQueryError):
            self.method("5,5,5,5")

    def test_list_of_1_raises(self):
        with pytest.raises(GeocoderQueryError):
            self.method([5])

    # TODO maybe raise for `[5, 5]` too?

    def test_list_of_3_raises(self):
        with pytest.raises(GeocoderQueryError):
            self.method([5, 5, 5])

    def test_list_of_4_raises(self):
        with pytest.raises(GeocoderQueryError):
            self.method([5, 5, 5, 5])

    def test_list_of_5_raises(self):
        with pytest.raises(GeocoderQueryError):
            self.method([5, 5, 5, 5, 5])

    def test_points(self):
        bbox = self.method([Point(50, 160), Point(30, 170)])
        assert bbox == "30.0,160.0,50.0,170.0"

    def test_lists(self):
        bbox = self.method([[50, 160], [30, 170]])
        assert bbox == "30.0,160.0,50.0,170.0"
        bbox = self.method([["50", "160"], ["30", "170"]])
        assert bbox == "30.0,160.0,50.0,170.0"

    def test_strings(self):
        bbox = self.method(["50, 160", "30,170"])
        assert bbox == "30.0,160.0,50.0,170.0"

    def test_output_format(self):
        bbox = self.method([Point(50, 160), Point(30, 170)],
                           " %(lon2)s|%(lat2)s -- %(lat1)s|%(lon1)s ")
        assert bbox == " 170.0|50.0 -- 30.0|160.0 "


@pytest.mark.parametrize("adapter_factory", [DummySyncAdapter, DummyAsyncAdapter])
async def test_synchronize_decorator_sync_simple(adapter_factory):
    geocoder = Geocoder(adapter_factory=adapter_factory)
    calls = []

    @_synchronized
    def f(self, one, *, two):
        calls.append((one, two))
        return 42

    res = f(geocoder, 1, two=2)
    if adapter_factory is DummyAsyncAdapter:
        res = await res
    assert 42 == res
    assert calls == [(1, 2)]


async def test_synchronize_decorator_async_simple():
    geocoder = Geocoder(adapter_factory=DummyAsyncAdapter)
    calls = []

    @_synchronized
    def f(self, one, *, two):
        async def coro():
            calls.append((one, two))
            return 42
        return coro()

    assert 42 == await f(geocoder, 1, two=2)
    assert calls == [(1, 2)]


@pytest.mark.parametrize("adapter_factory", [DummySyncAdapter, DummyAsyncAdapter])
async def test_synchronize_decorator_sync_exception(adapter_factory):
    geocoder = Geocoder(adapter_factory=adapter_factory)

    @_synchronized
    def f(self, one, *, two):
        raise RuntimeError("test")

    with pytest.raises(RuntimeError):
        res = f(geocoder, 1, two=2)
        if adapter_factory is DummyAsyncAdapter:
            await res


async def test_synchronize_decorator_async_exception():
    geocoder = Geocoder(adapter_factory=DummyAsyncAdapter)

    @_synchronized
    def f(self, one, *, two):
        async def coro():
            raise RuntimeError("test")
        return coro()

    with pytest.raises(RuntimeError):
        await f(geocoder, 1, two=2)


@pytest.mark.parametrize("adapter_factory", [DummySyncAdapter, DummyAsyncAdapter])
async def test_synchronize_decorator_sync_reentrance(adapter_factory):
    calls = []

    class DummyGeocoder(Geocoder):
        @_synchronized
        def f(self, i=0):
            calls.append(i)
            if len(calls) < 5:
                return self.f(i + 1)
            return 42

    geocoder = DummyGeocoder(adapter_factory=adapter_factory)

    res = geocoder.f()
    if adapter_factory is DummyAsyncAdapter:
        res = await res
    assert 42 == res
    assert calls == list(range(5))


async def test_synchronize_decorator_async_reentrance():
    calls = []

    class DummyGeocoder(Geocoder):
        @_synchronized
        def f(self, i=0):
            async def coro():
                calls.append(i)
                if len(calls) < 5:
                    return await self.f(i + 1)
                return 42
            return coro()

    geocoder = DummyGeocoder(adapter_factory=DummyAsyncAdapter)

    assert 42 == await geocoder.f()
    assert calls == list(range(5))



import collections.abc
from functools import partial
from urllib.parse import urlencode

from geopy.geocoders.base import DEFAULT_SENTINEL, Geocoder
from geopy.location import Location
from geopy.util import logger

__all__ = ("Photon", )


class Photon(Geocoder):
    """Geocoder using Photon geocoding service (data based on OpenStreetMap
    and service provided by Komoot on https://photon.komoot.de).

    Documentation at:
        https://github.com/komoot/photon

    Photon/Komoot geocoder aims to let you `search as you type with
    OpenStreetMap`. No API Key is needed by this platform.
    """

    geocode_path = '/api'
    reverse_path = '/reverse'

    def __init__(
            self,
            *,
            scheme=None,
            timeout=DEFAULT_SENTINEL,
            proxies=DEFAULT_SENTINEL,
            domain='photon.komoot.de',
            user_agent=None,
            ssl_context=DEFAULT_SENTINEL,
            adapter_factory=None
    ):
        """

        :param str scheme:
            See :attr:`geopy.geocoders.options.default_scheme`.

        :param int timeout:
            See :attr:`geopy.geocoders.options.default_timeout`.

        :param dict proxies:
            See :attr:`geopy.geocoders.options.default_proxies`.

        :param str domain: Should be the localized Photon domain to
            connect to. The default is ``'photon.komoot.de'``, but you
            can change it to a domain of your own.

        :param str user_agent:
            See :attr:`geopy.geocoders.options.default_user_agent`.

        :type ssl_context: :class:`ssl.SSLContext`
        :param ssl_context:
            See :attr:`geopy.geocoders.options.default_ssl_context`.

        :param callable adapter_factory:
            See :attr:`geopy.geocoders.options.default_adapter_factory`.

            .. versionadded:: 2.0
        """
        super().__init__(
            scheme=scheme,
            timeout=timeout,
            proxies=proxies,
            user_agent=user_agent,
            ssl_context=ssl_context,
            adapter_factory=adapter_factory,
        )
        self.domain = domain.strip('/')
        self.api = "%s://%s%s" % (self.scheme, self.domain, self.geocode_path)
        self.reverse_api = "%s://%s%s" % (self.scheme, self.domain, self.reverse_path)

    def geocode(
            self,
            query,
            *,
            exactly_one=True,
            timeout=DEFAULT_SENTINEL,
            location_bias=None,
            language=False,
            limit=None,
            osm_tag=None
    ):
        """
        Return a location point by address.

        :param str query: The address or query you wish to geocode.

        :param bool exactly_one: Return one result or a list of results, if
            available.

        :param int timeout: Time, in seconds, to wait for the geocoding service
            to respond before raising a :class:`geopy.exc.GeocoderTimedOut`
            exception. Set this only if you wish to override, on this call
            only, the value set during the geocoder's initialization.

        :param location_bias: The coordinates to used as location bias.

        :param str language: Preferred language in which to return results.

        :param int limit: Limit the number of returned results, defaults to no
            limit.

        :param osm_tag: The expression to filter (include/exclude) by key and/
            or value, str as ``'key:value'`` or list/set of str if multiple
            filters are required as ``['key:!val', '!key', ':!value']``.
        :type osm_tag: str or list or set

        :rtype: ``None``, :class:`geopy.location.Location` or a list of them, if
            ``exactly_one=False``.

        """
        params = {
            'q': query
        }
        if limit:
            params['limit'] = int(limit)
        if exactly_one:
            params['limit'] = 1
        if language:
            params['lang'] = language
        if location_bias:
            try:
                lat, lon = self._coerce_point_to_string(location_bias).split(',')
                params['lon'] = lon
                params['lat'] = lat
            except ValueError:
                raise ValueError(("Location bias must be a"
                                  " coordinate pair or Point"))
        if osm_tag:
            if isinstance(osm_tag, str):
                params['osm_tag'] = [osm_tag]
            else:
                if not isinstance(osm_tag, collections.abc.Iterable):
                    raise ValueError(
                        "osm_tag must be a string or "
                        "an iterable of strings"
                    )
                params['osm_tag'] = list(osm_tag)
        url = "?".join((self.api, urlencode(params, doseq=True)))
        logger.debug("%s.geocode: %s", self.__class__.__name__, url)
        callback = partial(self._parse_json, exactly_one=exactly_one)
        return self._call_geocoder(url, callback, timeout=timeout)

    def reverse(
            self,
            query,
            *,
            exactly_one=True,
            timeout=DEFAULT_SENTINEL,
            language=False,
            limit=None
    ):
        """
        Return an address by location point.

        :param query: The coordinates for which you wish to obtain the
            closest human-readable addresses.
        :type query: :class:`geopy.point.Point`, list or tuple of ``(latitude,
            longitude)``, or string as ``"%(latitude)s, %(longitude)s"``.

        :param bool exactly_one: Return one result or a list of results, if
            available.

        :param int timeout: Time, in seconds, to wait for the geocoding service
            to respond before raising a :class:`geopy.exc.GeocoderTimedOut`
            exception. Set this only if you wish to override, on this call
            only, the value set during the geocoder's initialization.

        :param str language: Preferred language in which to return results.

        :param int limit: Limit the number of returned results, defaults to no
            limit.

        :rtype: ``None``, :class:`geopy.location.Location` or a list of them, if
            ``exactly_one=False``.
        """
        try:
            lat, lon = self._coerce_point_to_string(query).split(',')
        except ValueError:
            raise ValueError("Must be a coordinate pair or Point")
        params = {
            'lat': lat,
            'lon': lon,
        }
        if limit:
            params['limit'] = int(limit)
        if exactly_one:
            params['limit'] = 1
        if language:
            params['lang'] = language
        url = "?".join((self.reverse_api, urlencode(params)))
        logger.debug("%s.reverse: %s", self.__class__.__name__, url)
        callback = partial(self._parse_json, exactly_one=exactly_one)
        return self._call_geocoder(url, callback, timeout=timeout)

    def _parse_json(self, resources, exactly_one=True):
        """
        Parse display name, latitude, and longitude from a JSON response.
        """
        if not len(resources['features']):  # pragma: no cover
            return None
        if exactly_one:
            return self._parse_resource(resources['features'][0])
        else:
            return [self._parse_resource(resource) for resource
                    in resources['features']]

    def _parse_resource(self, resource):
        # Return location and coordinates tuple from dict.
        name_elements = ['name', 'housenumber', 'street',
                         'postcode', 'street', 'city',
                         'state', 'country']
        name = [resource['properties'].get(k) for k
                in name_elements if resource['properties'].get(k)]
        location = ', '.join(name)

        latitude = resource['geometry']['coordinates'][1] or None
        longitude = resource['geometry']['coordinates'][0] or None
        if latitude and longitude:
            latitude = float(latitude)
            longitude = float(longitude)

        return Location(location, (latitude, longitude), resource)



from unittest.mock import patch

import pytest

import geopy.exc
import geopy.geocoders
from geopy.geocoders import What3Words
from test.geocoders.util import BaseTestGeocoder, env


class TestUnitWhat3Words:
    dummy_api_key = 'DUMMYKEY1234'

    async def test_user_agent_custom(self):
        geocoder = What3Words(
            api_key=self.dummy_api_key,
            user_agent='my_user_agent/1.0'
        )
        assert geocoder.headers['User-Agent'] == 'my_user_agent/1.0'

    @patch.object(geopy.geocoders.options, 'default_scheme', 'http')
    def test_default_scheme_is_ignored(self):
        geocoder = What3Words(api_key=self.dummy_api_key)
        assert geocoder.scheme == 'https'


@pytest.mark.skipif(
    not bool(env.get('WHAT3WORDS_KEY')),
    reason="No WHAT3WORDS_KEY env variable set"
)
class TestWhat3Words(BaseTestGeocoder):
    @classmethod
    def make_geocoder(cls, **kwargs):
        return What3Words(
            env['WHAT3WORDS_KEY'],
            timeout=3,
            **kwargs
        )

    async def test_geocode(self):
        await self.geocode_run(
            {"query": "piped.gains.jangle"},
            {"latitude": 53.037611, "longitude": 11.565012},
        )

    async def test_reverse(self):
        await self.reverse_run(
            {"query": "53.037611,11.565012", "lang": 'DE'},
            {"address": 'fortschrittliche.voll.schnitt'},
        )

    async def test_unicode_query(self):
        await self.geocode_run(
            {
                "query": (
                    "\u0070\u0069\u0070\u0065\u0064\u002e\u0067"
                    "\u0061\u0069\u006e\u0073\u002e\u006a\u0061"
                    "\u006e\u0067\u006c\u0065"
                )
            },
            {"latitude": 53.037611, "longitude": 11.565012},
        )

    async def test_empty_response(self):
        with pytest.raises(geopy.exc.GeocoderQueryError):
            await self.geocode_run(
                {"query": "definitely.not.existingiswearrrr"},
                {},
                expect_failure=True
            )

    async def test_not_exactly_one(self):
        await self.geocode_run(
            {"query": "piped.gains.jangle", "exactly_one": False},
            {"latitude": 53.037611, "longitude": 11.565012},
        )
        await self.reverse_run(
            {"query": (53.037611, 11.565012), "exactly_one": False},
            {"address": "piped.gains.jangle"},
        )

    async def test_result_language(self):
        await self.geocode_run(
            {"query": "piped.gains.jangle", "lang": 'DE'},
            {"address": 'fortschrittliche.voll.schnitt'},
        )

    async def test_check_query(self):
        result_check_threeword_query = self.geocoder._check_query(
            "\u0066\u0061\u0068\u0072\u0070\u0072"
            "\u0065\u0069\u0073\u002e\u006c\u00fc"
            "\u0067\u006e\u0065\u0072\u002e\u006b"
            "\u0075\u0074\u0073\u0063\u0068\u0065"
        )

        assert result_check_threeword_query
